@@ -4,6 +4,8 @@ toolchest_client.tools.test
 
 This is the test implementation of the Tool class.
 """
+from toolchest_client.files import OutputType
+
 from . import Tool
 
 
@@ -21,4 +23,9 @@ class Test(Tool):
             inputs=inputs,
             min_inputs=1,
             max_inputs=100,  # this limit is completely arbitrary
+            max_input_bytes_per_file=256 * 1024 * 1024 * 1024,
+            parallel_enabled=False,
+            output_type=OutputType.GZ_TAR,
+            output_is_directory=True,
+            output_names=["test_output.txt"],
         )
