@@ -19,10 +19,11 @@ from toolchest_client.api.exceptions import ToolchestException
 from toolchest_client.api.status import ThreadStatus
 from toolchest_client.api.query import Query
 from toolchest_client.files import files_in_path, split_file_by_lines, sanity_check, check_file_size,\
-    split_paired_files_by_lines, compress_files_in_path, OutputType
-from toolchest_client.tools.arg_whitelist import ARGUMENT_WHITELIST, VARIABLE_ARGS
+    split_paired_files_by_lines, compress_files_in_path, OutputType, assert_exists
+from toolchest_client.files.s3 import inputs_are_in_s3
+from toolchest_client.tools.tool_args import TOOL_ARG_LISTS, VARIABLE_ARGS
 
-FOUR_POINT_FIVE_GIGABYTES = 4.5 * 1024 * 1024 * 1024
+FOUR_POINT_FIVE_GIGABYTES = int(4.5 * 1024 * 1024 * 1024)
 
 
 class Tool:
@@ -30,9 +31,11 @@ class Tool:
                  output_path, inputs, min_inputs, max_inputs,
                  database_name=None, database_version=None,
                  input_prefix_mapping=None, parallel_enabled=False,
-                 max_input_bytes_per_node=FOUR_POINT_FIVE_GIGABYTES,
+                 max_input_bytes_per_file=FOUR_POINT_FIVE_GIGABYTES,
+                 max_input_bytes_per_file_parallel=FOUR_POINT_FIVE_GIGABYTES,
                  group_paired_ends=False, compress_inputs=False,
-                 output_type=OutputType.FLAT_TEXT, output_is_directory=False):
+                 output_type=OutputType.FLAT_TEXT, output_is_directory=False,
+                 output_names=None):
         self.tool_name = tool_name
         self.tool_version = tool_version
         self.tool_args = tool_args
@@ -52,13 +55,16 @@ class Tool:
         self.database_name = database_name
         self.database_version = database_version
         self.parallel_enabled = parallel_enabled
+        self.output_validation_enabled = True
         self.group_paired_ends = group_paired_ends
         self.compress_inputs = compress_inputs
-        self.max_input_bytes_per_node = max_input_bytes_per_node
+        self.max_input_bytes_per_file = max_input_bytes_per_file
+        self.max_input_bytes_per_file_parallel = max_input_bytes_per_file_parallel
         self.query_threads = []
         self.query_thread_statuses = dict()
         self.terminating = False
-        self.output_type = output_type
+        self.output_type = output_type or OutputType.FLAT_TEXT
+        self.output_names = output_names or []
         signal.signal(signal.SIGTERM, self._handle_termination)
         signal.signal(signal.SIGINT, self._handle_termination)
 
@@ -80,8 +86,97 @@ class Tool:
             raise ValueError(f"Too many input files submitted. "
                              f"Maximum is {self.max_inputs}, {self.num_input_files} found.")
 
+    def _validate_tool_args(self):
+        """
+        Validates and sanitizes user-provided custom tool_args.
+
+        This is processed as an argument whitelist; argument tags
+        are kept only if they appear as an accepted tag. If an argument
+        is not on the whitelist, an error is thrown.
+
+        If a dangerous argument – one that changes the function or
+        structure of the Toolchest run – is found, complexity is
+        reduced (no validation, no parallelization) and a warning
+        is shown.
+        """
+
+        whitelist = TOOL_ARG_LISTS[self.tool_name]["whitelist"]  # all tools have a whitelist
+        dangerlist = TOOL_ARG_LISTS[self.tool_name].get("dangerlist", [])  # some tools have a dangerlist
+        blacklist = TOOL_ARG_LISTS[self.tool_name].get("blacklist", [])  # some tools have a blacklist
+
+        sanitized_args = []  # arguments that are explicitly allowed
+        unknown_args = []  # all arguments that were not included
+        blacklisted_args = []  # arguments that are known to not work
+        dangerous_args = []  # arguments that significantly change the function of the program
+
+        num_args_remaining_after_tag = 0
+
+        # process arguments individually
+        for arg in self.tool_args.split():
+            # the minimal_tag for "--arg=a" is "--arg"
+            # this allows matching args assigned via an "=" to match on the whitelist
+            minimal_tag = re.sub(r"([^=]+)(=[^\s]+)", rf"\1", arg)
+            tag_in_whitelist = minimal_tag in whitelist
+            if num_args_remaining_after_tag == 0 and tag_in_whitelist:
+                # if the arg is a tag in the whitelist, add it
+                sanitized_args.append(arg)
+                num_args_remaining_after_tag = whitelist[minimal_tag]
+                # if the arg is a tag in the dangerlist, note so can adjust accordingly
+                if minimal_tag in dangerlist:
+                    dangerous_args.append(arg)
+            elif num_args_remaining_after_tag == VARIABLE_ARGS:
+                # if previous tag has additional args (unknown/variable amount),
+                # append args until another tag is found
+                # TODO: filter out non-escaped bash command-line characters
+                # TODO: handle variable arguments better; this allows passing of undesired args
+                sanitized_args.append(arg)
+                if tag_in_whitelist:
+                    num_args_remaining_after_tag = whitelist[minimal_tag]
+            elif num_args_remaining_after_tag > 0:
+                # append remaining args if previous tag has additional args
+                # TODO: filter out non-escaped bash command-line characters
+                sanitized_args.append(arg)
+                num_args_remaining_after_tag -= 1
+            else:
+                # Instead of stopping and throwing an error immediately, we wait and collect all undesired args
+                if minimal_tag in blacklist:
+                    blacklisted_args.append(arg)
+                else:
+                    unknown_args.append(arg)
+
+        if unknown_args or blacklisted_args:
+            print("Non-allowed arguments found in tool_args:")
+            print(
+                f"Blacklisted arguments (these are known to cause Toolchest to fail): \
+{blacklisted_args if blacklisted_args else '(none)'}"
+            )
+            print(
+                f"Unknown arguments (these are not yet validated for use with Toolchest – please contact us!): \
+{unknown_args if unknown_args else '(none)'}"
+            )
+            raise ValueError("Unknown or blacklisted arguments present in tool_args. See above for details.")
+
+        if dangerous_args:
+            print("WARNING: dangerous arguments found in tool_args. This disables validation and parallelization!")
+            print(f"Dangerous arguments: {dangerous_args}")
+            # Disable parallelization, validation, and revert to plain compressed output
+            self.output_validation_enabled = False
+            self.parallel_enabled = False
+            self.output_is_directory = True
+            self.output_type = OutputType.GZ_TAR
+
+        sanitized_args = " ".join(sanitized_args)
+        if sanitized_args != self.tool_args:
+            self.tool_args = sanitized_args
+        print("Processing tool_args as:")
+        pretty_print_args = self.tool_args if self.tool_args else "(no tool_args set)"
+        print(f"\t{pretty_print_args}")
+
     def _validate_args(self):
-        """Validates args set by tools."""
+        # Perform a deep tool_args validation
+        # This has to happen before checking the input args, as in some cases parallelization is disabled and
+        # expected input / output values may change.
+        self._validate_tool_args()
 
         if self.inputs is None:
             raise ValueError("No input provided.")
@@ -97,65 +192,41 @@ class Tool:
         if self.output_is_directory and not os.path.isdir(self.output_path):
             raise ValueError(f"Output path must be a directory. It is currently {self.output_path}")
 
-        # Perform a deeper tool_args validation
-        self._validate_tool_args()
-
-    def _validate_tool_args(self):
-        """Validates and sanitizes user-provided custom tool_args.
-
-        Currently, this is processed as an argument whitelist; argument tags
-        are kept only if they appear as an accepted tag.
-        """
-
-        tool_arg_whitelist = ARGUMENT_WHITELIST[self.tool_name]
-
-        sanitized_args_list = []
-        num_args_remaining_after_tag = 0
-
-        # process arguments individually
-        for arg in self.tool_args.split():
-            # the minimal_tag for "--arg=a" is "--arg"
-            # this allows matching args assigned via an "=" to match on the whitelist
-            minimal_tag = re.sub(r"([^=]+)(=[^\s]+)", rf"\1", arg)
-            tag_in_whitelist = minimal_tag in tool_arg_whitelist
-            if num_args_remaining_after_tag == 0 and tag_in_whitelist:
-                # if the arg is a tag in the whitelist, add it
-                sanitized_args_list.append(arg)
-                num_args_remaining_after_tag = tool_arg_whitelist[minimal_tag]
-            elif num_args_remaining_after_tag == VARIABLE_ARGS:
-                # if previous tag has additional args (unknown/variable amount),
-                # append args until another tag is found
-                # TODO: filter out non-escaped bash command-line characters
-                # TODO: handle variable arguments better; this allows passing of undesired args
-                sanitized_args_list.append(arg)
-                if tag_in_whitelist:
-                    num_args_remaining_after_tag = tool_arg_whitelist[minimal_tag]
-            elif num_args_remaining_after_tag > 0:
-                # append remaining args if previous tag has additional args
-                # TODO: filter out non-escaped bash command-line characters
-                sanitized_args_list.append(arg)
-                num_args_remaining_after_tag -= 1
-            # if no tag found, skip args until a tag is found
-
-        sanitized_args = " ".join(sanitized_args_list)
-        if sanitized_args != self.tool_args:
-            self.tool_args = sanitized_args
-        print("Processing tool_args as:")
-        pretty_print_args = self.tool_args if self.tool_args else "(no tool_args set)"
-        print(f"\t{pretty_print_args}")
-
     def _merge_outputs(self, output_file_paths):
         """Merges output files for parallel runs."""
         raise NotImplementedError(f"Merging outputs not enabled for this tool {self.tool_name}")
 
+    def _warn_if_outputs_exist(self):
+        """Warns if default output files already exist in the output directory"""
+        for file_path in self.output_names + ["output", "output.tar.gz"]:
+            joined_file_path = os.path.join(self.output_path, file_path)
+            if os.path.exists(joined_file_path):
+                print(f"WARNING: {joined_file_path} already exists and will be overwritten")
+
     def _preflight(self):
         """Generic preflight check. Tools can have more specific implementations."""
         # Validate Toolchest auth key.
         validate_key()
 
+        if self.output_is_directory:
+            if os.path.exists(self.output_path):
+                if os.path.isfile(self.output_path):
+                    raise ValueError(
+                        f"{self.output_path} is a file. Please pass a directory instead of an output file."
+                    )
+            else:
+                os.makedirs(self.output_path)
+
+        self._warn_if_outputs_exist()
+
     def _postflight(self):
         """Generic postflight check. Tools can have more specific implementations."""
-        sanity_check(self.output_path)
+        if self.output_validation_enabled:
+            for output_name in self.output_names:
+                output_file_path = f"{self.output_path}/{output_name}"
+                assert_exists(output_file_path, must_be_file=True)
+                if os.stat(output_file_path).st_size <= 5:
+                    raise ValueError(f"Output file at {output_file_path} is suspiciously small")
 
     def _system_supports_parallel_execution(self):
         """Checks if parallel execution is supported on the platform.
@@ -259,7 +330,7 @@ class Tool:
                 # Arbitrary parallelization – assume only one input file which is to be split
                 adjusted_input_file_paths = split_file_by_lines(
                     input_file_path=self.input_files[0],
-                    max_bytes=self.max_input_bytes_per_node,
+                    max_bytes=self.max_input_bytes_per_file_parallel,
                 )
                 for _, file_path in adjusted_input_file_paths:
                     # This is assuming only one input file per parallel run.
@@ -269,15 +340,15 @@ class Tool:
                 # Grouped parallelization. Right now, this only supports grouping by R1/R2 for paired-end inputs
                 input_file_paths_pairs = split_paired_files_by_lines(
                     input_file_paths=self.input_files,
-                    max_bytes=self.max_input_bytes_per_node,
+                    max_bytes=self.max_input_bytes_per_file_parallel,
                 )
                 for input_file_path_pair in input_file_paths_pairs:
                     yield input_file_path_pair
 
         else:
-            # Make sure we're below plan/multi-part limit for non-splittable files
+            # Make sure we're below tool limit for non-splittable files
             for file_path in self.input_files:
-                check_file_size(file_path, max_size_bytes=FOUR_POINT_FIVE_GIGABYTES)
+                check_file_size(file_path, max_size_bytes=self.max_input_bytes_per_file)
             # Note that for a tool like Unicycler, this would look like:
             # [["r1.fastq", "r2.fastq", "unassembled.fasta"]]
             # As there are multiple input files required for the job
@@ -287,18 +358,20 @@ class Tool:
         """Constructs and runs a Toolchest query."""
         print("Beginning Toolchest analysis run.")
 
-        self._preflight()
-
         # todo: better propagate and handle errors for parallel runs
         self._validate_args()
+
+        # Preflight check should occur after validating args, as validation may affect the preflight check
+        self._preflight()
+
         # Prepare input files (expand paths, compress, etc)
         self._prepare_inputs()
 
         print(f"Found {self.num_input_files} files to upload.")
 
         should_run_in_parallel = self.parallel_enabled \
-            and self.group_paired_ends or self.num_input_files == 1 \
-            and check_file_size(self.input_files[0]) > self.max_input_bytes_per_node \
+            and not any(inputs_are_in_s3(self.input_files)) \
+            and (self.group_paired_ends or self.num_input_files == 1) \
             and self._system_supports_parallel_execution()
 
         jobs = self._generate_jobs(should_run_in_parallel)
@@ -364,4 +437,3 @@ class Tool:
             self._postflight()
 
         print("Analysis run complete!")
-
